#include <windows.h>
#include <stdio.h>
#include <tchar.h>

#if defined(ASSERT)
#define assert(e) ASSERT(e)
#else  // defined(ASSERT)
#include <assert.h>
#define ASSERT(e) assert(e)
#endif  // defined(ASSERT)

#if !defined(VERIFY)
#if defined(DEBUG) || defined(_DEBUG)
#define VERIFY(e) ASSERT(e)
#else  // defined(DEBUG) || defined(_DEBUG)
#define VERIFY(e) ((void)(e))
#endif  // defined(DEBUG) || defined(_DEBUG)
#endif  // !defined(VERIFY)

#define ARRAY_LENGTH(x) (sizeof(x)/sizeof((x)[0]))

#if defined(UNICODE) || defined(_UNICODE)
#define TString WideString
#else  // defined(UNICODE) || defined(_UNICODE)
#define TString AnsiString
#endif  // defined(UNICODE) || defined(_UNICODE)
#define AnsiString(x) (x)
#define WideString(x) (x)

int _sntprintfz(TCHAR* const psDestination, const unsigned int uiDestinationLength, const TCHAR* const psFormat, ...)
	{
	int iReturn = 0;
	if(0 < uiDestinationLength)
		{
		va_list va;
		va_start(va, psFormat);
		iReturn = _vsntprintf(psDestination, uiDestinationLength-1, psFormat, va);
		va_end(va);
		if(0 < iReturn)
			{
			psDestination[uiDestinationLength-1] = TEXT('\0');
			iReturn = _tcslen(psDestination);
			}
		else
			{
			psDestination[0] = TEXT('\0');
			iReturn = 0;
			}
		}
	return iReturn;
	}

#if defined(DEBUG) || defined(_DEBUG)
#if !defined(DEBUG)
#define DEBUG
#endif // !defined(DEBUG)
#endif // defined(DEBUG) || defined(_DEBUG)

#define MAX_STRING_LENGTH (256)

const TCHAR* ptsIgnoreWindowClasses[] =
	{
	TEXT("Progman"),
	};

int StringCompareFunction(const void* pv0, const void* pv1)
	{
	const TCHAR* const pts0 = *(TCHAR**)pv0;
	const TCHAR* const pts1 = *(TCHAR**)pv1;
	return _tcscmp(pts0, pts1);
	}

struct CaptureEnumWindowsProcData
	{
	bool* pbSuccess;
	FILE* pfOutput;
	};

BOOL CALLBACK CaptureEnumWindowsProc(HWND hWnd, LPARAM lParam)
	{
	ASSERT(CaptureEnumWindowsProc == (WNDENUMPROC)CaptureEnumWindowsProc);
	const CaptureEnumWindowsProcData* const pcewpd = (CaptureEnumWindowsProcData*)lParam;
	ASSERT((0 != pcewpd) && (0 != pcewpd->pbSuccess) && (0 != pcewpd->pfOutput));
	FILE* const pfOutput = pcewpd->pfOutput;
	ASSERT(0 != pfOutput);
	if(IsWindowVisible(hWnd))
		{
		TCHAR ptsWindowTitle[MAX_STRING_LENGTH];
		if(0 == GetWindowText(hWnd, ptsWindowTitle, ARRAY_LENGTH(ptsWindowTitle)))
			{
			ptsWindowTitle[0] = TEXT('\0');
			}
		TCHAR ptsWindowClass[MAX_STRING_LENGTH];
		if(0 == GetClassName(hWnd, ptsWindowClass, ARRAY_LENGTH(ptsWindowClass)))
			{
			ptsWindowClass[0] = TEXT('\0');
			}
		const TCHAR* const ptsWindowClassPointer = ptsWindowClass;
		const TCHAR* const* const pptsWindowClass = &ptsWindowClassPointer;
		WINDOWPLACEMENT wp;
		wp.length = sizeof(wp);
		if(
		   (TEXT('\0') != ptsWindowTitle[0]) &&
		   (TEXT('\0') != ptsWindowClass[0]) &&
		   (0 == bsearch(pptsWindowClass, ptsIgnoreWindowClasses, ARRAY_LENGTH(ptsIgnoreWindowClasses), sizeof(ptsIgnoreWindowClasses[0]), StringCompareFunction)) &&
		   GetWindowPlacement(hWnd, &wp))
			{
#if defined(DEBUG)
			TCHAR ptsDebugString[MAX_STRING_LENGTH];
			_sntprintfz(ptsDebugString, ARRAY_LENGTH(ptsDebugString), TEXT("hWnd=0x%08x Class=\"%s\" Name=\"%s\" showCmd=0x%08x ptMinPosition=(%d,%d) ptMaxPosition=(%d,%d) rcNormalPosition=[(%d,%d)-(%d,%d)]\n"), hWnd, ptsWindowClass, ptsWindowTitle, wp.showCmd, wp.ptMinPosition.x, wp.ptMinPosition.y, wp.ptMaxPosition.x, wp.ptMaxPosition.y, wp.rcNormalPosition.left, wp.rcNormalPosition.top, wp.rcNormalPosition.right, wp.rcNormalPosition.bottom);
			OutputDebugString(ptsDebugString);
#endif // defined(DEBUG)
			const int iftprintfResult = _ftprintf(pfOutput, TEXT("\"%s\" \"%s\" [(%d,%d)-(%d,%d)]\n"), AnsiString(ptsWindowClass), AnsiString(ptsWindowTitle), wp.rcNormalPosition.left, wp.rcNormalPosition.top, wp.rcNormalPosition.right, wp.rcNormalPosition.bottom);
			if(iftprintfResult < 0)
				{
				*(pcewpd->pbSuccess) = false;
				}
			}
		}
	return (pcewpd->pbSuccess ? TRUE : FALSE);
	}

bool CaptureWindowPositions(FILE* const pfOutput)
	{
#if defined(DEBUG)
	// Ensure that ptsIgnoreWindowClasses array is sorted
	for(int i = 0 ; i < ARRAY_LENGTH(ptsIgnoreWindowClasses)-1 ; i++)
		{
		ASSERT(_tcscmp(ptsIgnoreWindowClasses[i], ptsIgnoreWindowClasses[i+1]) < 0);
		}
#endif // defined(DEBUG)
	bool bSuccess = true;
	CaptureEnumWindowsProcData cewpd;
	cewpd.pbSuccess = &bSuccess;
	cewpd.pfOutput = pfOutput;
	VERIFY(EnumWindows(CaptureEnumWindowsProc, (LPARAM)(&cewpd)));
	return bSuccess;
	}

void SkipSpace(bool &rbSuccess, const TCHAR* &rptc)
	{
	while(rbSuccess && _istspace(*rptc))
		{
		rptc++;
		}
	}

void ReadString(bool &rbSuccess, const TCHAR* &rptc, TCHAR* const ptsString, const size_t stStringSize)
	{
	TCHAR* ptcString = ptsString;
	rbSuccess &= (TEXT('\"') == *rptc);
	rptc++;
	while(rbSuccess && (TEXT('\0') != *rptc) && (TEXT('\"') != *rptc) && ((unsigned int)(ptcString-ptsString) < (stStringSize-1)))
		{
		*ptcString = *rptc;
		ptcString++;
		rptc++;
		}
	*ptcString = TEXT('\0');
	rbSuccess &= (TEXT('\"') == *rptc);
	rptc++;
	}

void ReadNumber(bool &rbSuccess, const TCHAR* &rptc, long& rlNumber)
	{
	rlNumber = 0;
	bool bNegative = false;
	while(rbSuccess && (TEXT('\0') != *rptc) && (_istdigit(*rptc) || ((TEXT('-') == *rptc))))
		{
		if(TEXT('-') == *rptc)
			{
			rbSuccess &= (!bNegative && (0 == rlNumber));
			bNegative = true;
			}
		else
			{
			rlNumber = (10*rlNumber)+(*rptc-TEXT('0'));
			}
		rptc++;
		}
	if(rbSuccess && bNegative)
		{
		rlNumber = -rlNumber;
		}
	}

void VerifyChar(bool &rbSuccess, const TCHAR* &rptc, const TCHAR c)
	{
	rbSuccess &= (c == *rptc);
	rptc++;
	}

bool GetWindowInformationFromString(const TCHAR* const ptsInput, TCHAR* const ptsWindowClass, const size_t stWindowClassSize, TCHAR* const ptsWindowTitle, const size_t stWindowTitleSize, WINDOWPLACEMENT* const pwp)
	{
	ASSERT((0 != ptsInput) && (0 != pwp));
	bool bSuccess = true;
	const TCHAR* ptc = ptsInput;
	SkipSpace (bSuccess, ptc);
	ReadString(bSuccess, ptc, ptsWindowClass, stWindowClassSize);
	SkipSpace (bSuccess, ptc);
	ReadString(bSuccess, ptc, ptsWindowTitle, stWindowTitleSize);
	SkipSpace (bSuccess, ptc);
	VerifyChar(bSuccess, ptc, TEXT('['));
	SkipSpace (bSuccess, ptc);
	VerifyChar(bSuccess, ptc, TEXT('('));
	SkipSpace (bSuccess, ptc);
	ReadNumber(bSuccess, ptc, pwp->rcNormalPosition.left);
	SkipSpace (bSuccess, ptc);
	VerifyChar(bSuccess, ptc, TEXT(','));
	SkipSpace (bSuccess, ptc);
	ReadNumber(bSuccess, ptc, pwp->rcNormalPosition.top);
	SkipSpace (bSuccess, ptc);
	VerifyChar(bSuccess, ptc, TEXT(')'));
	SkipSpace (bSuccess, ptc);
	VerifyChar(bSuccess, ptc, TEXT('-'));
	SkipSpace (bSuccess, ptc);
	VerifyChar(bSuccess, ptc, TEXT('('));
	SkipSpace (bSuccess, ptc);
	ReadNumber(bSuccess, ptc, pwp->rcNormalPosition.right);
	SkipSpace (bSuccess, ptc);
	VerifyChar(bSuccess, ptc, TEXT(','));
	SkipSpace (bSuccess, ptc);
	ReadNumber(bSuccess, ptc, pwp->rcNormalPosition.bottom);
	SkipSpace (bSuccess, ptc);
	VerifyChar(bSuccess, ptc, TEXT(')'));
	SkipSpace (bSuccess, ptc);
	VerifyChar(bSuccess, ptc, TEXT(']'));
	return bSuccess;
	}

struct RestoreEnumWindowsProcData
	{
	const TCHAR* ptsWindowClass;
	const TCHAR* ptsWindowTitle;
	const WINDOWPLACEMENT* pwp;
	};

BOOL CALLBACK RestoreEnumWindowsProc(HWND hWnd, LPARAM lParam)
	{
	ASSERT(CaptureEnumWindowsProc == (WNDENUMPROC)CaptureEnumWindowsProc);
	const RestoreEnumWindowsProcData* const prewpd = (RestoreEnumWindowsProcData*)lParam;
	ASSERT((0 != prewpd) && (0 != prewpd->ptsWindowClass) && (0 != prewpd->ptsWindowTitle) && (0 != prewpd->pwp));
	TCHAR ptsWindowClass[MAX_STRING_LENGTH];
	TCHAR ptsWindowTitle[MAX_STRING_LENGTH];
	if((0 < GetClassName(hWnd, ptsWindowClass, ARRAY_LENGTH(ptsWindowClass))) && (0 == _tcscmp(ptsWindowClass, prewpd->ptsWindowClass)) && 
	   (0 < GetWindowText(hWnd, ptsWindowTitle, ARRAY_LENGTH(ptsWindowTitle))) && (0 != _tcsstr(ptsWindowTitle, prewpd->ptsWindowTitle)))
		{
		VERIFY(SetWindowPlacement(hWnd, prewpd->pwp));
		}
	return TRUE;
	}

bool RestoreWindowPositions(FILE* const pfInput)
	{
	bool bSuccess = true;
	WINDOWPLACEMENT wp;
	ZeroMemory(&wp, sizeof(wp));
	wp.length = sizeof(wp);
	// wp.flags = WPF_ASYNCWINDOWPLACEMENT;
	wp.showCmd = SW_SHOWNA;
	RestoreEnumWindowsProcData rewpd;
	rewpd.ptsWindowClass = 0;
	rewpd.ptsWindowTitle = 0;
	rewpd.pwp = &wp;
	while(bSuccess && !feof(pfInput))
		{
		TCHAR pcsInputLine[MAX_STRING_LENGTH];
		if(0 != _fgetts(pcsInputLine, ARRAY_LENGTH(pcsInputLine), pfInput))
			{
			if((TEXT('\0') != pcsInputLine[0]) && (TEXT('\r') != pcsInputLine[0]) && (TEXT('\n') != pcsInputLine[0]) && (TEXT('#') != pcsInputLine[0]))
				{
				TCHAR ptsWindowClass[MAX_STRING_LENGTH];
				TCHAR ptsWindowTitle[MAX_STRING_LENGTH];
				if(GetWindowInformationFromString(TString(pcsInputLine), ptsWindowClass, ARRAY_LENGTH(ptsWindowClass), ptsWindowTitle, ARRAY_LENGTH(ptsWindowTitle), &wp))
					{
#if defined(DEBUG)
					TCHAR ptsDebugString[MAX_STRING_LENGTH];
					_sntprintfz(ptsDebugString, ARRAY_LENGTH(ptsDebugString), TEXT("Class=\"%s\" rcNormalPosition=[(%d,%d)-(%d,%d)]\n"), ptsWindowClass, wp.rcNormalPosition.left, wp.rcNormalPosition.top, wp.rcNormalPosition.right, wp.rcNormalPosition.bottom);
					OutputDebugString(ptsDebugString);
#endif // defined(DEBUG)
					rewpd.ptsWindowClass = ptsWindowClass;
					rewpd.ptsWindowTitle = ptsWindowTitle;
					VERIFY(EnumWindows(RestoreEnumWindowsProc, (LPARAM)(&rewpd)));
					}
				else
					{
					bSuccess = false;
					}
				}
			}
		else
			{
			if(!feof(pfInput))
				{
				bSuccess = false;
				}
			}
		}
	return bSuccess;
	}

int main(const int argc, const char** argv)
	{
	if(1 < argc)
		{
		if((0 == _strcmpi(argv[1], "-h")) || (0 == _strcmpi(argv[1], "/h")) ||
		   (0 == _strcmpi(argv[1], "-help")) || (0 == _strcmpi(argv[1], "/help")) ||
		   (0 == _strcmpi(argv[1], "-?")) || (0 == _strcmpi(argv[1], "/?")))
			{
			_ftprintf(stdout,
				TEXT("\n")
				TEXT("WindowPlacementTool\n")
				TEXT("Copyright (c) 2000 by David Anson (DavidAns@Microsoft.com)\n")
				TEXT("\n")
				TEXT("[-h | -help | -?]\n")
				TEXT("\tThis help screen\n")
				TEXT("\n")
				TEXT("[-c | -capture] [Capture_file_name.txt]\n")
				TEXT("\tCapture the current window positions to a file (or standard output if\n")
				TEXT("\tno file name is given)\n")
				TEXT("\n")
				TEXT("[Restore_file_name_1.txt] [Restore_file_name_2.txt] ...\n")
				TEXT("\tRestore the window positions from the data previously captured in the\n")
				TEXT("\tspecified file(s) (or standard input if no file name is given)\n")
				);
			}
		else if((0 == _strcmpi(argv[1], "-c")) || (0 == _strcmpi(argv[1], "/c")) ||
		        (0 == _strcmpi(argv[1], "-capture")) || (0 == _strcmpi(argv[1], "/capture")))
			{
			if(2 == argc)
				{
				VERIFY(CaptureWindowPositions(stdout));
				}
			else if(3 == argc)
				{
				const char* const pcsFileName = argv[2];
				FILE* const pfOutput = fopen(pcsFileName, "w");
				if(0 != pfOutput)
					{
					if(!CaptureWindowPositions(pfOutput))
						{
						_ftprintf(stderr, TEXT("\nError: Unable to write to file: \"%s\"\n"), pcsFileName);
						}
					VERIFY(0 == fclose(pfOutput));
					}
				else
					{
					_ftprintf(stderr, TEXT("\nError: Unable to open file for write: \"%s\"\n"), pcsFileName);
					}
				}
			else
				{
				_ftprintf(stderr, TEXT("\nError: Must specify only one output file\n"));
				}
			}
		else
			{
			for(int i = 1 ; i < argc ; i++)
				{
				const char* const pcsFileName = argv[i];
				FILE* const pfInput = fopen(pcsFileName, "r");
				if(0 != pfInput)
					{
					if(!RestoreWindowPositions(pfInput))
						{
						_ftprintf(stderr, TEXT("\nError: Unable to read file: \"%s\"\n"), pcsFileName);
						}
					VERIFY(0 == fclose(pfInput));
					}
				else
					{
					_ftprintf(stderr, TEXT("\nError: Unable to open file for read: \"%s\"\n"), pcsFileName);
					}
				}
			}
		}
	else
		{
		VERIFY(RestoreWindowPositions(stdin));
		}
	return 0;
	}
